// le90m.h - LE90M SUMMA synchonization with U70 module by Karpekov & Matveev 
//
// Nov-2007 by Solo
//
#ifndef _LE90M_H
#define _LE90M_H

#include "le20c.h"

#define LE90M_BS       0
#define LE90M_GATE     1
#define LE90M_ADDR     2
#define LE90M_PARAM    3

#define LE90M_DISABLE  0
#define LE90M_TSSCOPE  1
#define LE90M_TBSGATE  2
#define LE90M_TLED     3
#define LE90M_TGATEES  4
#define LE90M_TINHIBIT 5
#define LE90M_TESSTART 6
#define LE90M_FLED     7
#define LE90M_FTEST    8

// all times are in 0.1 sec units in BCD format
// for example 0x16 = 1.6 seconds
// frequencies are specified as a divisor from base frequency
// all functions are using already converted values
typedef struct
{
  short tbs;       // time between start and begin spill
  short tgate;     // width of the gate
  short disable;   // disable manual programming 
  short tsscope;   // time between start and scope start
  short tbsgate;   // time between begin spill and gate
  short tled;      // width of led pulses window
  short tgatees;   // time between end of gate and end spill
  short tinhibit;  // width of inhibit signal after end spill
  short tesstart;  // time between end spill and next start
  short fled;      // led pulses frequency divisor, 1=100Hz, 0=none
  short ftest;     // test pulses frequency divisor, 1=100kHz, 0=none
} LE90MDATA;

class LE90M
{

 public:

  LE90M (LE20C* host, int number)
  {
    n = number;
    h = host;
  }

  ~LE90M () {};

  int number() { return n; }
  LE20C * host() { return h; }

  short bcd2dec(short w) { return (w>>4)*10+(w&0xf); }
  short dec2bcd(short w) { short h=w/10; return (h<<4)+(w-h*10); }

  short tbs() { return bcd2dec(h->read(n,LE90M_BS)); }
  short tgate() { return bcd2dec(h->read(n,LE90M_GATE)); } 

  void write_tbs(short data) { h->write(n,LE90M_BS,dec2bcd(data)); }
  void write_tgate(short data) { h->write(n,LE90M_GATE,dec2bcd(data)); }

  void write (int addr, short data) {
    h->write(n,LE90M_ADDR,addr);
    h->write(n,LE90M_PARAM,dec2bcd(data));
  }

  void write (LE90MDATA* data) {
    write_tbs(data->tbs);
    write_tgate(data->tgate);
    short *d=&data->tsscope;
    for(int i=1; i<9; i++) write(i,*d++);
    write(0,data->disable); // any write clears this flag, should be the last one
  }

  short read (int addr) {
      h->write(n,LE90M_ADDR,addr);
      return bcd2dec(h->read(n,LE90M_PARAM));
  }

  void read (LE90MDATA* data) {
    data->tbs=tbs();
    data->tgate=tgate();
    short *d=&data->disable;
    for(int i=0; i<9; i++) *d++=read(i);
  }

  void enable() { write(0,0); }
  void disable() { write(0,1); }

  // save configuration
  int save(const char *filename, LE90MDATA *data=NULL) {
    unsigned short *s;
    LE90MDATA rdata;

    if (data==NULL) {
      read (&rdata);
      data=&rdata;
    }

    FILE *fp = fopen(filename,"w");
    if(fp==NULL) {
      printf("le90m::save cannot open file %s for writing!\n",filename);
      return -1;
    }

    s = (unsigned short *)data;
    for(int i=0; i<11; i++) fprintf(fp, "%hd\n", *(s++));

    fclose(fp);
    return 0;
  }

  int load(const char *filename, LE90MDATA *data=NULL) {
    unsigned short *s;
    LE90MDATA wdata;

    FILE *fp = fopen(filename,"r");
    if(fp==NULL) {
      printf("le90m::load cannot open file %s for reading!\n",filename);
      return -1;
    }

    if(data!=NULL) s=(unsigned short *)data;
    else s=(unsigned short *)&wdata;

    for(int i=0; i<11; i++) fscanf(fp, "%hd", s++);
    
    fclose(fp);

    if(data==NULL) write(&wdata);

    return 0;
  }

  void print  (LE90MDATA *data=NULL) {
    LE90MDATA rdata;

    if (data==NULL) {
      read (&rdata);
      data=&rdata;
    }

    printf("LE90M module %d contents:\n",n);
    printf("tbs      = %.1f s\n",data->tbs/10.0);
    printf("tgate    = %.1f s\n",data->tgate/10.0);
    printf("disable  = %s \n",data->disable==0?"no":"yes");
    printf("tsscope  = %.1f s\n",data->tsscope/10.0);
    printf("tbsgate  = %.1f s\n",data->tbsgate/10.0);
    printf("tled     = %.1f s\n",data->tled/10.0);
    printf("tgatees  = %.1f s\n",data->tgatees/10.0);
    printf("tinhibit = %.1f s\n",data->tinhibit/10.0);
    printf("tesstart = %.1f s\n",data->tesstart/10.0);
    printf("fled     = %.1f Hz \n",data->fled==0?0:100.0/data->fled);
    printf("ftest    = %.1f kHz \n",data->ftest==0?0:100.0/data->ftest);
  }

 private:
  LE20C* h; // QBUS-SUMMA host
  int n;    // position in crate

};

# endif
