// gams.cc - main GAMS routine to update pedestals
//
// Nov-2004 by Solo, from Serge sources
// Nov-2006 by Solo, upgraded for LE-71
// Nov-2007 by Solo, upgraded LE20I to LE83
// Nov-2008 by Solo, add DIM support for data and control

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>

#include "qbus.h"  // QBUS interface
#include "miss.h"  // MISS abstract interface
//#include "le20i.h" // MISS-QBUS interface
#include "le83.h"  // MISS-QBUS interface
#include "le20c.h" // SUMMA-QBUS interface
#include "le24.h"  // ADC controller LE24, LE66 or LE68
#include "le71.h"  // LE-71 ADC
#include "pulse.h" // pulser
#include "bs225.h" // latch
#include "go.h"    // GAMS movement
#include "ether.h" // Calibration connection

#include "mysql.h" // MySQL database

#include "dis.hxx" // DIM services

#define ADC_MAX_MODULES 3
#define ADC_MAX_CRATES  6
#define MEM_MAX_CRATES  3
#define ZERO_MAX_PEDS 100 


// Global variables
MYSQL mysql;    // mysql connection handle
MYSQL_RES *res; // result handle
MYSQL_ROW row;  // row handle

bool Running=false; // running flag

int Cycles=200;   // number of pedestal cycles
int Debug=1;      // debug level
int Dbase=0;      // database flag
int Calib=0;      // calibration flag
int Dim=0;        // DIM service flag
int Thr=4;        // ped threshold for LE68
int Run=0;        // run number
int Log=0;        // GAMS movement log file

float MeanThr=0.5; // threshold for database update

int Spill=0;    // spill number
int Crates=0;   // number of ADC crates
int Mems=0;     // number of MEM crates
int Zero=0;     // number of zero pedestals

QBUS *qb[2]; // PC-QBUS interfaces
LE20C *su;   // Summa controller
BS225 *bs;   // end of spill latch
PULSE *pu;   // pulser to trigger pedestal readout
GO    *go;   // mover

LE83 *mi[ADC_MAX_CRATES]; // 6 MISS ADC crates
LE83 *me[MEM_MAX_CRATES]; // 3 MISS memory crates

int zero_peds[ZERO_MAX_PEDS];

int peds_cycles[ADC_MAX_CRATES];
short peds_read[LE71_MAX_CHANNELS];

long peds_sum[ADC_MAX_CRATES][LE71_MAX_CHANNELS];
long rms_sum[ADC_MAX_CRATES][LE71_MAX_CHANNELS];

short peds_write[ADC_MAX_CRATES][LE71_MAX_CHANNELS];
double mean[ADC_MAX_CRATES][LE71_MAX_CHANNELS];
double rms[ADC_MAX_CRATES][LE71_MAX_CHANNELS];
double mean_old[ADC_MAX_CRATES][LE71_MAX_CHANNELS];

// DIM services
DimService *srvGamsSpill;
DimCommand *cmdGamsMove;
DimCommand *cmdGamsStop;
DimService *srvGamsStatus;
DimService *srvGamsXmoving;
DimService *srcGamsYmoving;
DimService *srvGamsXY;
DimService *srvGamsCRQ;

XY dimXY;
CRQ dimCRQ;
char dimStatus[80];

FILE *logfile=NULL;

// execute command
int execcmd (unsigned char *buffer)
{
    int cmd,len;
    CRQ coord;
    CRQ curr;
    int ax;

    XY xy;

    if(Dim) {
      cmd = 1;
      len = 8;
      sscanf((const char *)buffer,"%02d%02d%1d",&coord.column,&coord.row,&coord.quarter);
    }
    else {
      cmd = buffer[0];
      len = buffer[1];
      coord.column  = buffer[2];
      coord.row = buffer[4];
      coord.quarter = buffer[6];
    }

    // check parameters
    if(coord.column<1 || coord.column>29) {
      printf("Parameter column %d is out of [1-29] range!\n",coord.column);
      return -1;
    }
    if(coord.row <1 || coord.row>22) {
      printf("Parameter row %d is out of [1-22] range!\n",coord.row);
      return -1;
    }
    if(coord.quarter<0 || coord.quarter>4) {
      printf("Parameter quarter %d is out of [1-4] range!\n",coord.quarter);
      return -1;
    }

    printf("Executing command %d, len = %d, coords: %d %d %d",
	   cmd,len,coord.column,coord.row,coord.quarter);

    go->get_XY (&xy);
    go->xy2crq (&xy, &curr);

    ax = XY_ENABLE;
    if ((coord.column == curr.column) and
       ((coord.quarter == curr.quarter) or
       ((coord.quarter == 1) and (curr.quarter == 2)) or
       ((coord.quarter == 2) and (curr.quarter == 1)) or
       ((coord.quarter == 3) and (curr.quarter == 4)) or
       ((coord.quarter == 4) and (curr.quarter == 3))))
      ax -= X_ENABLE;

    if ((coord.row == curr.row) and
       ((coord.quarter == curr.quarter) or
       ((coord.quarter == 1) and (curr.quarter == 3)) or
       ((coord.quarter == 3) and (curr.quarter == 1)) or
       ((coord.quarter == 2) and (curr.quarter == 4)) or
       ((coord.quarter == 4) and (curr.quarter == 2))))
      ax -= Y_ENABLE;

    go->crq2xy (&coord, &xy);
    printf(" : %d %d\n",xy.X,xy.Y);

    if(Dim) {
      strcpy(dimStatus,"moving");
      srvGamsStatus->updateService();
    }

    go->move(&xy, ax);

    printf("==========> Command done\n");

    if(Dim) {
      go->get_XY (&dimXY);
      go->xy2crq (&dimXY, &dimCRQ);
      strcpy(dimStatus,"stopped");
      srvGamsStatus->updateService();
      srvGamsXY->updateService();
      srvGamsCRQ->updateService();
    }

    return(0);
}

// read configuration, create devices
int gams_config(char *config_path)
{
  FILE *cfg;
  char s[133];
  char name[80]="";
  char rest[80]="";
  int value=0;
  int index=0;

  printf("Configuring...\n");

  cfg = fopen(config_path,"r");
  if(cfg==NULL) {
    printf("Cannot open file %s for reading!\n", config_path);
    return -1;
  }

  printf("Processing config file %s\n",config_path);

  // parse config file
  while(fgets(s,sizeof(s),cfg)!=NULL) {
    // skip comments
    if(strlen(s)<2 || s[0]=='#' || s[0]==' ' || s[0]=='!' || s[0]=='/') continue;

    name[0]=0; rest[0]=0; value=0;
    sscanf(s,"%s %d %s",name,&value,rest);
    if(Debug>0) printf("%s %d %s\n",name,value,rest);

    // Variables
    if(strcmp(name,"cycles")==0) Cycles=value;
    else if(strcmp(name,"debug")==0) Debug=value;
    else if(strcmp(name,"dbase")==0) Dbase=value;
    else if(strcmp(name,"calib")==0) Calib=value;
    else if(strcmp(name,"dim")==0) Dim=value;
    else if(strcmp(name,"thr")==0) Thr=value;
    else if(strcmp(name,"run")==0) Run=value;
    // log file
    else if(strcmp(name,"log")==0) {
      if(value==1) {
        logfile=fopen(rest,"a");
        if(logfile!=NULL) Log=1;
	//t=time(0);
	//fprintf(logfile,"\n\t-----------\t%s\t-------------\n",ctime(&t));
	else {
	  printf("Failed to open log file %s!\n",rest);
	  return -1;
        }
      }
    }
    // QBUS
    else if(strcmp(name,"qbus")==0) {
      if(value<0 || value>1) {
	printf("qbus index %d is out of [0,1] range!\n",value);
	if(Debug<2) return -1;
      }
      else {
	qb[value] = new QBUS(rest);
	if(qb[value]->node()<0) {
	  printf("QBUS device %s not found! Check qbus driver...\n",rest); 
	  return -1;
	}
	qb[value]->reset();
      }
    }
    // SUMMA crate
    else if(strcmp(name,"summa")==0) {
      if(value<0 || value>16) {
	printf("summa crate %d is out of [0,16] range!\n",value);
	if(Debug<2) return -1;
      }
      else {
	sscanf(rest,"%d",&index);
	if(index<0 || index>1) printf("qbus index %d is out of [0,1] range!\n",index);
	else su = new LE20C(qb[index], value);
      }
    }
    // SUMMA modules
    else if(strcmp(name,"latch")==0) {
      if(value<1 || value>24) printf("summa N %d is out of [1,24] range!\n",value);
      else bs = new BS225(su, value);
    }
    else if(strcmp(name,"pulse")==0) {
      if(value<1 || value>24) printf("summa N %d is out of [1,24] range!\n",value);
      else pu = new PULSE(su, value);
    }
    else if(strcmp(name,"go")==0) {
      if(value<1 || value>24) printf("summa N %d is out of [1,24] range!\n",value);
      else {
	sscanf(rest,"%d",&index);
	if(index<1 || index>24) printf("summa N %d is out of [1,24] range!\n",index);
	else go = new GO(su, value, index, logfile);
      }
    }
    // MISS ADC crates
    else if(strcmp(name,"adc")==0) {
      if (Crates>=ADC_MAX_CRATES) {
	printf("Number of ADC crates is out of [1-6] range!\n");
	continue;
      }
      if(value<0 || value>15) printf("crate %d is out of [0,15] range!\n",value);
      else {
	sscanf(rest,"%d",&index);
	if(index<0 || index>1) printf("qbus index %d is out of [0,1] range!\n",index);
	else {
	  mi[Crates] = new LE83(qb[index], value);
	  mi[Crates]->au();
	  mi[Crates]->clear();
	  if(mi[Crates]->status()<0) {
	    printf("MISS crate %d not found! Check crate power...\n",mi[Crates]->crate);
	    mi[Crates]=NULL;    // remove crate from config
	    qb[index]->reset(); // need to reset QBUS after bad access
	    sleep(1);
	  }
	}
      }
      Crates++;
    }
    // MISS memory crates
    else if(strcmp(name,"mem")==0) {
      if (Mems>=MEM_MAX_CRATES) {
	printf("Number of MEM crates is out of [0-3] range!\n");
	continue;
      }
      if(value<0 || value>15) printf("crate %d is out of [0,15] range!\n",value);
      else {
	sscanf(rest,"%d",&index);
	if(index<0 || index>1) printf("qbus index %d is out of [0,1] range!\n",index);
	else {
	  me[Mems] = new LE83(qb[index], value);
	  me[Mems]->au();
	  me[Mems]->clear();
	  if((me[Mems]->status()&0x07ff)<0) {
	    printf("MISS crate %d not found! Check crate power...\n",me[Mems]->crate);
	    me[Mems]=NULL;      // remove crate from config
	    qb[index]->reset(); // need to reset QBUS after bad access
	    sleep(1);
	  }
	}
      }
      Mems++;
    }
    // zero pedestals
    else if(strcmp(name,"zero")==0) {
      if(Zero>=ZERO_MAX_PEDS) {
	printf("Number of ZERO pedestals is out of [0-100] range!\n");
	continue;
      }
      int crate=value/10000;
      int module=(value-crate*10000)/100;
      int channel=value-crate*10000-module*100;

      // invert groups of 24 channels for LE-71
      int section=channel/24;
      channel=channel-section*24;
      channel=(24-channel)+section*24;

      if(crate<0 || crate>5) printf("Crate %d is out of [0,5] range!\n",crate);
      else if (module<1 || module>4) printf("Module %d is out of [1,4] range!\n",module);
      else if (channel<0 || channel>95) printf("Channel %d is out of [0,95] range!\n",channel);
      else zero_peds[Zero++]=value;      
    }

    // Unknown
    else printf("Unknown command!\n");
  }

  // finally connect to MySQL database
  if(Dbase>0) {
    if (! mysql_connect(&mysql,"pcbech.ihep.su","roger","gregor")) {
      printf("Cannot connect to MySQL database: %s\n",mysql_error(&mysql));
      return -1;
    }

    if ( mysql_select_db(&mysql,"gams") ) {
      printf("Cannot select database: %s\n",mysql_error(&mysql));
      return -1;
    }
  }

  // clear mean values, to force DB update
  for(int i=0; i<ADC_MAX_CRATES; i++) {
    for(int j=0; j<LE71_MAX_CHANNELS; j++) {
      mean_old[i][j]=-10.0;
    }
  }

  return 0;
}

// wait for end of spill
int wait_endspill()
{
  while(bs->event()==0) {
    sleep(1);
  }  

  // TODO - add timeout and QBUS reset
  return 1;
}

// prepare for spill or for pedestals
int get_ready(int ped=0) 
{

  // clear QBUS
  qb[0]->clear();
  qb[1]->clear();

  // send common clear
  //pu->oc();

  // clear memory crates
  for(int i=0; i<Mems; i++) {
    if(me[i]!=NULL) {
      me[i]->au();
      if (ped==1) me[i]->clear();
    }
  }

  // clear ADC crates
  for(int i=0; i<Crates; i++) {
    if(mi[i]!=NULL) {
      mi[i]->au();
      if (ped==1) mi[i]->clear();
    }
  }

  // restart end spill latch
  if(ped==0) bs->restart(1);

  return 0;
}

// get pedestals
int get_ped_le71()
{
  struct timeval tv0, tv1;
  LE71 *adc;

  // clear sums
  for(int i=0; i<Crates; i++) {
    peds_cycles[i]=0;
    for(int j=0; j<LE71_MAX_CHANNELS; j++) {
      peds_sum[i][j]=0;
      rms_sum[i][j]=0;
    }
    // clear peds in LE71
    if(mi[i]!=NULL) {
      for(int k=0; k<4; k++) {
	LE71 *adc = new LE71(mi[i],k+1);
	adc->clear();
      }
    }
  }

  gettimeofday(&tv0,0);  
  for(int cycle=1; cycle<=Cycles; cycle++) {

    // pulse pedestal trigger
    pu->ped();
    
    // loop over crates
    for(int i=0; i<Crates; i++) {
      int n=0;
      int channels=LE71_MAX_CHANNELS;
      int min_channels=96;

      // if(i==5) channels=14*16;
      // if (i==5) channels = 13*16;

      if(mi[i]!=NULL) {
	//n=mi[i]->sdata(LE71_MAX_CHANNELS,peds_read);
	n=mi[i]->sdata_fast16(LE71_MAX_CHANNELS,peds_read);
	//n=mi[i]->sdata_fast(LE71_MAX_CHANNELS,peds_read);
	//	if(n<channels) {
	if ( n<min_channels ) {
	  if( Debug>0 ) printf("Crate %d serial data read error: read %d words out of %d\n",mi[i]->crate,n,channels);
	}
	else {
	  // add sum
	  for(int j=0; j<n; j++) {
	    peds_sum[i][j]+=peds_read[j];
	    rms_sum[i][j] +=peds_read[j]*peds_read[j];
	  }
	  peds_cycles[i]++;
	}
	mi[i]->clear();
      }// end if crate
    }// end for

    if(Debug>0 && cycle%10==0) { printf("%d\r",cycle); fflush(stdout); }
  }
  gettimeofday(&tv1,0);

  // restore pedestals
  for(int i=0; i<Crates; i++) {
    if(mi[i]!=NULL /*&& mi[i]->crate!=5*/ ) {  
      for(int j=0; j<4; j++) {
	adc = new LE71(mi[i],j+1);
	adc->write(&peds_write[i][j*96]);
      }
    }
  }
  
  // now calculate mean and RMS (<x^2>-<x>^2)
  for(int i=0; i<Crates; i++) {
    if(mi[i]!=NULL /*&& mi[i]->crate!=5*/) {
      for(int j=0; j<LE71_MAX_CHANNELS; j++) {
	mean[i][j] = (1.0*peds_sum[i][j])/peds_cycles[i];
	rms[i][j]  = sqrt(fabs((1.0*rms_sum[i][j])/peds_cycles[i]-mean[i][j]*mean[i][j]));
	peds_write[i][j] = (short)((mean[i][j] + 0.5));
      }
    }
  }

  if(Debug>0) {
    // print readout statistics
    double usec = (tv1.tv_sec-tv0.tv_sec)*1000000.0 + tv1.tv_usec-tv0.tv_usec;
    printf("%.3f sec/spill, %.1f usec/word \n",usec/1000000.0, usec/(Cycles*LE71_MAX_CHANNELS*Crates));
  }

  return 0;
}

// Set pedestals in ADC memory
int set_ped_le71() 
{
  int tot_refreshed=0;

  // add threshold to the pedestal
  for(int i=0; i<Crates; i++) {
    for(int j=0; j<LE71_MAX_CHANNELS; j++) {
      if(peds_write[i][j]<4090) peds_write[i][j]+=Thr;
    }
  }

  // zero some special pedestals coded like CCMMAA - Crate Module Address
  for(int i=0; i<Zero; i++) {
    int crate=zero_peds[i]/10000;
    int module=(zero_peds[i]-crate*10000)/100;
    int channel=zero_peds[i]-crate*10000-module*100;
    for(int j=0; j<ADC_MAX_CRATES; j++) {
      if(mi[j]!=NULL && mi[j]->crate==crate) peds_write[j][(module-1)*96+channel]=0;
    }
  }

  for(int i=0; i<ADC_MAX_CRATES; i++) {
    if(mi[i]!=NULL /*&& mi[i]->crate!=5*/) {
      int refreshed=0;
      LE71* adc;

      // refresh pedestals
      for(int j=0; j<4; j++) {
	adc = new LE71(mi[i],j+1);	
	refreshed += adc->refresh(&peds_write[i][j*96]);
      }

      if (Debug>0 && refreshed>0) 
	printf("Refreshed %d channels in crate %d\n",refreshed,mi[i]->crate);
      tot_refreshed += refreshed;
    }
  }

  if (Debug>0) printf("Refreshed %d channels in total\n",tot_refreshed);

  // massive refresh problem
  if (Debug>1 && tot_refreshed>1000) {
    
  } 

  return tot_refreshed;
}

// save pedestals in the database
int save_ped_le71()
{
  char qhead[]="(`n`,`time`,`spill`,`crate`,`module`,`chan`,`mean`,`rms`)";

  char query[1000];

  struct timeval tv0, tv1;
  int updated=0;

  gettimeofday(&tv0,0);
  // loop over crates
  for(int i=0; i<Crates; i++) {
    if(mi[i]!=NULL) {
      // loop over channels
      for(int c=0; c<LE71_MAX_CHANNELS; c++) {
	  if( fabs(mean[i][c]-mean_old[i][c])>MeanThr || Spill%100==0 ) {
	    int module=c/96;
	    int chan=c-module*96;
	    sprintf(query,"INSERT INTO run%d %s VALUES ('',%s,%d,%d,%d,%d,%f,%f);",
		    Run,qhead,"NOW()",Spill,mi[i]->crate+1,module+1,chan+1,mean[i][c],rms[i][c]);
	    if (mysql_query(&mysql, query)!=0) 
	      printf("Error saving to the database: %s\n",mysql_error(&mysql));
	    mean_old[i][c]=mean[i][c]; // save new mean
	    updated++;
	  }
      } // end channels
    }
  } // end crates
  gettimeofday(&tv1,0);

  if (Debug>0) printf("Updated %d channels\n",updated); 

  double secs = tv1.tv_sec-tv0.tv_sec + (tv1.tv_usec-tv0.tv_usec)/1000000.0;
  if (Debug>0) printf("Data saved to db in %.3f secs\n",secs);

  return 0;
}


// main routine
int main(int argc, char **argv)
{
  int i;
  time_t t;
  int len,tmo,machine,s;
  char config_path[256];
  unsigned char buffer[BUFLEN], answer[BUFLEN];

  printf("*** GAMS control program V2.3 by Solo ***\n");
  
  // configure
  if (argc>1) strncpy(config_path,argv[1],256);
  else strncpy(config_path,"gams.cfg",256);

  if (gams_config(config_path)<0) return -1;

  // Dim init
  if(Dim>0) {
    srvGamsSpill = new DimService("BEC_GAMS/SPILL",Spill);
    srvGamsStatus = new DimService("BEC_GAMS/STATUS",dimStatus);
    srvGamsXY = new DimService("BEC_GAMS/XY","I:2",&dimXY,sizeof(dimXY));
    srvGamsCRQ = new DimService("BEC_GAMS/CRQ","I:3",&dimCRQ,sizeof(dimCRQ));
    if (Calib>0) {
      cmdGamsMove = new DimCommand("BEC_GAMS/MOVE","C");
      cmdGamsStop = new DimCommand("BEC_GAMS/STOP","C");
    }
    DimServer::start("BEC_GAMS");
  }

  // print current position
  printf("\nCurrent GAMS position:\n");
  go->print();

  // update position and status
  if(Dim>0) {
    go->get_XY (&dimXY);
    go->xy2crq (&dimXY, &dimCRQ);
    strcpy(dimStatus,"stopped");
    srvGamsStatus->updateService();
    srvGamsXY->updateService();
    srvGamsCRQ->updateService();
  } 

  // connect to remote control (old VMS system) 
  if (Calib>0 && Dim==0) {
    printf("Connecting to remote control...\n");
    if (sconnect(0,0) <0 ) return -1;

    machine=2;
    printf("Opening satellite box...\n");
    if ( openbox(machine) ) return disconnect();
  }

  // prepare for run
  get_ready();

  if(Debug>0) printf("Waiting for end of spill...\n");
  Running=true;

  while(Running) {
    // wait for end of spill
    i=wait_endspill();
    if(i<1) continue; // timeout

    Spill++;
    time_t tt;
    tt=time(0);
    if(Debug>0) printf("Spill %d at %s",Spill,ctime(&tt));

    sleep(1); // wait some time after spill
    // get pedestals and calculate mean, RMS
    get_ready(1);
    get_ped_le71();

    // store new pedestals in controllers
    set_ped_le71();
    //    test_ped();
    get_ready();

    if (Dim>0) srvGamsSpill->updateService();

    // save pedestals in the database
    if(Dbase>0) save_ped_le71();

    // read and process remote command
    if ( Calib>0 ) {
      tmo = 1;

      if(Dim>0) {
	if(cmdGamsMove->getNext()) {
	  printf("Received command %s\n",cmdGamsMove->getString());
	  execcmd((unsigned char *)cmdGamsMove->getString());
	}
      }
      else {
	s = readcmd(buffer, BUFLEN, &tmo);
	if (s==0) {
	  printf("Received command %d\n",buffer[0]);
	  len = buffer[1]+2;
	  memcpy (answer, buffer, len);
	  tmo = 2;
	  answer[0] = 10;
	  s = writecmd (answer, len, &tmo);
	  s = execcmd(buffer);                      
	  tmo = 2;
	  answer[0] = 2;
	  s = writecmd (answer, len, &tmo);
	}
      }

      bs->restart(1);  // restore ES latch
    }

    // update position and status
    if(Dim>0) {
      go->get_XY (&dimXY);
      go->xy2crq (&dimXY, &dimCRQ);
      srvGamsXY->updateService();
      srvGamsCRQ->updateService();
    } 

  }

  mysql_close(&mysql);

  return 0;
}
