/*
  pci-qbus.c

  PCI-QBUS driver

  v1.0 Dec-2006 by Solo
  v1.1 Nov-2007 by Solo - new firmware of pci-qbus board
*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/uaccess.h> 

MODULE_AUTHOR("Oleg.Solovyanov@ihep.ru");
MODULE_DESCRIPTION("Driver for PCI-QBUS interface cards");
EXPORT_NO_SYMBOLS;

// global symbols
#define PCI_QBUS_VENDOR 0x1172
#define PCI_QBUS_ID          3

#define PCI_QBUS_MAXDEV      4
#define PCI_QBUS_IO_RANGE   32

#define PCI_QBUS_STATUS_PORT 0
#define PCI_QBUS_VECTOR_PORT 4
#define PCI_QBUS_ADDR_PORT   8
#define PCI_QBUS_DATA_PORT  12
#define PCI_QBUS_ADDW_PORT  16

#define PCI_QBUS_REG5_PORT  20
#define PCI_QBUS_REG6_PORT  24
#define PCI_QBUS_REG7_PORT  28

#define PCI_QBUS_READY       1
#define PCI_QBUS_TIMEOUT     2
#define PCI_QBUS_INTERRUPT   4

#define PCI_QBUS_DEVICE_NUM MINOR(filp->f_dentry->d_inode->i_rdev)

// global variables
static unsigned int pci_qbus_major=0;
static char name[PCI_QBUS_MAXDEV][4]={"pq0","pq1","pq2","pq3"};
static int io[PCI_QBUS_MAXDEV];
static int irq[PCI_QBUS_MAXDEV];
static int debug=0;

MODULE_PARM(debug,"i");
MODULE_PARM_DESC(debug,"debug level (default 0)");

// delay in usec using ISA port I/O
void pci_qbus_delay(int us) 
{ 
  int t;
  for (t=0; t<us; t++) outb(0,0x80); 
}

// interrupt service routine
void pci_qbus_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  if (debug>0) printk("pci-qbus.c: irq=%d interrupt from io=0x%x\n",irq,(int)dev_id);
}

// device open routine
int pci_qbus_open(struct inode *inode, struct file *filp)
{
  unsigned int port=io[PCI_QBUS_DEVICE_NUM]; // base port
  unsigned short w;

  int minor=MINOR(inode->i_rdev);
  if(debug>0) printk("pci-qbus.c: %s open(0x%x,%d)\n",name[minor],io[minor],irq[minor]);

  // check if card is accessible
  if(debug>0) {
    //    outw(0x5a5a, port+PCI_QBUS_REG5_PORT);
    //    w = inw(port+PCI_QBUS_REG5_PORT);
    //    if(w!=0x5a5a) printk("pci-qbus.c: %s error reg5 - write %x read %x base=%x\n",name[minor],0x5a5a,w,port);
  }

  return 0;
}

// device close routine
int pci_qbus_release(struct inode *inode, struct file *filp)
{
  int minor=MINOR(inode->i_rdev);
  if(debug>0) printk("pci-qbus.c: %s close\n",name[minor]);
  return 0;
}

// address setup
loff_t pci_qbus_llseek(struct file *filp, loff_t offset, int whence) 
{
  int port=io[PCI_QBUS_DEVICE_NUM]; // base port

  if(debug>0) printk("pci-qbus.c: %s lseek(0x%x,%d)\n",name[PCI_QBUS_DEVICE_NUM],((short)offset)&0xffff,whence);

  switch(whence) {
  case 0: { // SEEK_SET - set address
    filp->f_pos=offset;
    //    outw((short)filp->f_pos, port+PCI_QBUS_ADDR_PORT); 
    break; 
  }
  case 1: { // SEEK_CUR - clear error status
    outw(0, port+PCI_QBUS_VECTOR_PORT); 
    break;
  }    
  case 2: { // SEEK_END - device+branch reset
    outw(0, port+PCI_QBUS_STATUS_PORT);
    break;
  }
  default: break;
  }
  
  return filp->f_pos;
}

// read from device
ssize_t pci_qbus_read(struct file *filp, char *buf, size_t count, loff_t *offp)
{
  unsigned int port=io[PCI_QBUS_DEVICE_NUM]; // base port
  unsigned short w;

  outw(filp->f_pos, port+PCI_QBUS_ADDR_PORT); // write addr to perform cycle
  //  pci_qbus_delay(10);
  //  w = inw(port+PCI_QBUS_STATUS_PORT);         // check timeout
  while((w=inw(port+PCI_QBUS_STATUS_PORT))==0) ;  // wait for resonse
  if(w==1) {
    w = inw(port+PCI_QBUS_DATA_PORT);         // read data
    copy_to_user(buf, &w, 2);

    if(debug>0) printk("pci-qbus.c: %s read addr=0x%x data=0x%x\n",name[PCI_QBUS_DEVICE_NUM],(unsigned short)filp->f_pos,(unsigned short)w);
  
    return 2;
  }

  outw(0,PCI_QBUS_VECTOR_PORT); // clear timeout status  
  if(debug>0) printk("pci-qbus.c: %s read timeout!\n",name[PCI_QBUS_DEVICE_NUM]);
  return 0;
}

// write to device
ssize_t pci_qbus_write(struct file *filp, const char *buf, size_t count, loff_t *offp)
{
  unsigned int port=io[PCI_QBUS_DEVICE_NUM]; // base port
  unsigned short w;

  copy_from_user(&w, buf, 2);

  if(debug>0) printk("pci-qbus.c: %s write addr=0x%x data=0x%x\n",name[PCI_QBUS_DEVICE_NUM],(unsigned short)filp->f_pos,(unsigned short)w);

  outw(w, port+PCI_QBUS_DATA_PORT);           // set data word to write
//  pci_qbus_delay(10);
  outw(filp->f_pos, port+PCI_QBUS_ADDW_PORT); // write addr to perform cycle
  //  pci_qbus_delay(10);
  //  w = inw(port+PCI_QBUS_STATUS_PORT);         // check timeout
  while((w=inw(port+PCI_QBUS_STATUS_PORT))==0) ;  // wait for resonse
  if(w==1) {
    return 2;
  }

  outw(0,port+PCI_QBUS_VECTOR_PORT); // clear timeout status
  if(debug>0) printk("pci-qbus.c: %s write timeout 0x%x!\n",name[PCI_QBUS_DEVICE_NUM],w);
  return 0;
}

// available operations
struct file_operations pci_qbus_fops = {
  owner: THIS_MODULE,
  llseek: pci_qbus_llseek,
  read: pci_qbus_read,
  write: pci_qbus_write,
  //  ioctl: pci_qbus_ioctl,
  open: pci_qbus_open,
  release: pci_qbus_release,
};

int init_module(void)
{
  int i,err;
  printk("pci-qbus.c: v1.1 Nov-2007 by Solo\n");
  i=register_chrdev(0,"pq",&pci_qbus_fops);
  if(i<0) {
    printk("pci-qbus.c: can't get major number\n");
    return i;
  }
  pci_qbus_major=i;

  // clear tables
  for(i=0; i<PCI_QBUS_MAXDEV; i++) {
    io[i]=0;
    irq[i]=0;
  }

  // find available cards
  struct pci_dev *dev = NULL;
  for(i=0; i<PCI_QBUS_MAXDEV; i++) {
    dev = pci_find_device(PCI_QBUS_VENDOR, PCI_QBUS_ID, dev);
    if(!dev) break; // no more devices

    // enable device
    if ( (err=pci_enable_device(dev))<0 ) return err; 

    io[i]=pci_resource_start(dev, 0); // get the base I/O address
    irq[i]=dev->irq;
    printk("pci-qbus.c: found card io=0x%x irq=0x%x\n",io[i],irq[i]); 
  }

  // allocate hw resources
  for(i=0; i<PCI_QBUS_MAXDEV; i++) {
    if(io[i]>0) {
      if(!request_region(io[i],PCI_QBUS_IO_RANGE,name[i])<0) {
	printk("%s:cannot allocate PCI I/O port %d\n",name[i],io[i]);
	io[i]=0;
	irq[i]=0;
      }
      if(irq[i]>0) {
	if(request_irq(irq[i],pci_qbus_interrupt,SA_SHIRQ,name[i],&io[i])<0) {
	  printk("%s:cannot allocate interrupt line %d\n",name[i],irq[i]);
	  irq[i]=0;
	}
      }      
    }	
  }

  return 0;
}

void cleanup_module(void)
{
  int i;

  printk("pci-qbus.c: module unload\n");

  // release hw resources
  for(i=0; i<PCI_QBUS_MAXDEV; i++) {
    if(io[i]>0) {
      release_region(io[i],PCI_QBUS_IO_RANGE);
    }
    if(irq[i]>0) {
      free_irq(irq[i],NULL);
    }
  }

  unregister_chrdev(pci_qbus_major,"pq");
}

